/**
 * @typedef {Object} RgbColor
 * @property {number} r
 * @property {number} g
 * @property {number} b
 * @property {number} [a] 透明度，值置于 0 - 1 之间
 */

/**
 * @description 解析 rgb 字符串
 * @param {string} colorString 形如 #1e90ff 或者 rgba(0, 206, 209, 1) 这样的字符串
 * @returns {RgbColor | undefined}
 */
export function parseColor(colorString) {
    // 检查是否是十六进制颜色
    if (colorString.startsWith('#')) {
        let hex = colorString.slice(1);
        if (hex.length === 3) {
            hex = hex.split('').map(c => c + c).join('');
        }
        const r = parseInt(hex.slice(0, 2), 16);
        const g = parseInt(hex.slice(2, 4), 16);
        const b = parseInt(hex.slice(4, 6), 16);
        return { r, g, b };
    }
    // 检查是否是 RGBA 颜色
    else if (colorString.startsWith('rgba')) {
        const matches = colorString.match(/rgba\((\d+),\s*(\d+),\s*(\d+),\s*(\d+(\.\d+)?)\)/);
        if (matches) {
            const r = parseInt(matches[1], 10);
            const g = parseInt(matches[2], 10);
            const b = parseInt(matches[3], 10);
            const a = parseFloat(matches[4]);            
            return { r, g, b, a };
        }
    }
    // 检查是否是 RGB 颜色
    else if (colorString.startsWith('rgb')) {
        const matches = colorString.match(/rgb\((\d+),\s*(\d+),\s*(\d+)\)/);
        if (matches) {
            const r = parseInt(matches[1], 10);
            const g = parseInt(matches[2], 10);
            const b = parseInt(matches[3], 10);
            return { r, g, b };
        }
    }

    return undefined;
}

/**
 * @description 提升颜色的亮度
 * @param {RgbColor} rgb 
 * @param {number} percent 0 - 100 的数字，代表增强的亮度比例
 * @returns {RgbColor}
 */
export function increaseBrightness(rgb, percent) {
    // 确保 percent 在 0 到 100 之间
    percent = Math.max(0, Math.min(100, percent));

    // 计算每个颜色分量的增量
    const increment = (percent / 100) * 255;

    // 提升每个颜色分量的亮度
    const r = Math.min(255, Math.round(rgb.r + increment));
    const g = Math.min(255, Math.round(rgb.g + increment));
    const b = Math.min(255, Math.round(rgb.b + increment));

    return { r, g, b };
}


/**
 * @description 降低颜色的亮度
 * @param {RgbColor} rgb 
 * @param {number} percent 0 - 100 的数字，代表增强的亮度比例
 * @returns {RgbColor}
 */
export function lowerBrightness(rgb, percent) {
    // 确保 percent 在 0 到 100 之间
    percent = Math.max(0, Math.min(100, percent));

    // 计算每个颜色分量的增量
    const increment = (percent / 100) * 255;

    // 提升每个颜色分量的亮度
    const r = Math.min(255, Math.round(rgb.r - increment));
    const g = Math.min(255, Math.round(rgb.g - increment));
    const b = Math.min(255, Math.round(rgb.b - increment));

    return { r, g, b };
}


/**
 * @description gamma 修正
 * @param {number} c 颜色通道值，取值范围为 0 - 255
 */
function gammaCorrected(c) {
    c /= 255;
    return c <= 0.03928 ? c / 12.92 : Math.pow((c + 0.055) / 1.055, 2.4);
}

/**
 * @description 判断是否为亮色主题
 * @param {number} r 
 * @param {number} g 
 * @param {number} b 
 */
export function isLightColorTheme(r, g, b) {
    r = gammaCorrected(r);
    g = gammaCorrected(g);
    b = gammaCorrected(b);
    const luminance = 0.2126 * r + 0.7152 * g + 0.0722 * b;
    return luminance > 0.5;
}

/**
 * @description 导出为 rgb css 样式的字符串
 * @param {RgbColor} rgb 
 */
export function toRgbCssString(rgb) {
    const r = rgb.r;
    const g = rgb.g;
    const b = rgb.b;
    return `rgb(${r}, ${g}, ${b})`;
}

/**
 * @description 导出为 rgba css 样式的字符串
 * @param {RgbColor} rgb 
 */
export function toRgbaCssString(rgb) {
    const r = rgb.r;
    const g = rgb.g;
    const b = rgb.b;
    const a = rgb.a;
    return `rgb(${r}, ${g}, ${b}, ${a})`;
}

/**
 * @typedef ComputedColorOption
 * @property {string} BaseForegroundColorMacroName
 * @property {string} BaseBackgroundColorMacroName
 */

/**
 * @typedef GetColorOption
 * @property {'pdf' | 'svg'} mode
 */

export class MacroColor {
    /**
     * 
     * @param {ComputedColorOption} option 
     */
    constructor(option = {}) {
        this.option = option;
        this.rootStyles = getComputedStyle(document.documentElement);

        const foregroundColorString = this.rootStyles.getPropertyValue(option.BaseForegroundColorMacroName || '--foreground');
        const backgroundColorString = this.rootStyles.getPropertyValue(option.BaseBackgroundColorMacroName || '--background');
        const foregroundColor = parseColor(foregroundColorString);
        const backgroundColor = parseColor(backgroundColorString);
        // 通过背景颜色来判断当前颜色主题的类别
        const isLight = isLightColorTheme(backgroundColor.r, backgroundColor.g, backgroundColor.b);
        
        /**
         * @type {'light' | 'dark'}
         */
        this.theme = isLight ? 'light' : 'dark';

        this.foregroundColor = foregroundColor;
        this.backgroundColor = backgroundColor;
    }

    /**
     * 
     * @param {string} macroName 
     * @param {GetColorOption} option 
     */
    getColor(macroName, option = {}) {
        const theme = this.theme;
        const rootStyles = this.rootStyles;
        const mode = option.mode || 'svg';

        if (mode === 'svg') {
            // svg 模式下，导出的效果和 webview 渲染效果基本一致，直接导出即可
            return rootStyles.getPropertyValue(macroName);
        }
        
        // pdf 模式需要对 黑色主题 的几个特殊颜色进行一定的处理，并对所有透明颜色进行混合处理
        switch (macroName) {
        case '--foreground':
        case '--wire-color':
        case '--cross-dot-color':
            if (theme === 'dark') {
                return '#2D323B';
            }
        }

        const colorString = rootStyles.getPropertyValue(macroName);
        if (!colorString) {
            // 如果 macroName 不存在，那么此时的 colorString 是空的
            return colorString;
        }
        const color = parseColor(colorString);
        if (!color.a) {
            // 不具有透明通道，在 pdf 中渲染效果和 svg 中一致，直接返回即可
            return toRgbCssString(color);
        }

        // 透明度插值公式为 c = c_f * \alpha + c_b * (1 - \alpha)
        // c_f: 前景颜色
        // c_b: 背景颜色
        // alpha: 透明度
        // 该操作是 channel-wise 的
        const mixedBg = parseColor('#ffffff');
        const mixedColor = { r: 0, g: 0, b: 0 };
        const alpha = color.a;
        mixedColor.r = color.r * alpha + mixedBg.r * (1 - alpha);
        mixedColor.g = color.g * alpha + mixedBg.g * (1 - alpha);
        mixedColor.b = color.b * alpha + mixedBg.b * (1 - alpha);

        return toRgbCssString(mixedColor);
    }
}